# -*- coding: utf-8 -*-
import logging
from typing import TYPE_CHECKING

from IPython import get_ipython

from .trace_stmt import TraceStatement
from .trace_events import TraceEvent
from .trace_state import TraceState

if TYPE_CHECKING:
    from types import FrameType
    from ..safety import DependencySafety

logger = logging.getLogger(__name__)


def make_tracer(safety: 'DependencySafety'):
    if safety.trace_messages_enabled:
        logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.ERROR)

    def tracer(frame: 'FrameType', evt: str, _):
        # notebook cells have filenames that appear as '<ipython-input...>'
        if not frame.f_code.co_filename.startswith('<ipython-input'):
            return

        event = TraceEvent(evt)
        state = safety.trace_state  # we'll be using this a lot

        # IPython quirk -- every line in outer scope apparently wrapped in lambda
        # We want to skip the outer 'call' and 'return' for these
        if event == TraceEvent.call:
            state.call_depth += 1
            if state.call_depth == 1:
                return tracer

        if event == TraceEvent.return_:
            state.call_depth -= 1
            if state.call_depth == 0:
                return tracer

        cell_num, lineno = TraceState.get_position(frame)
        try:
            stmt_node = safety.statement_cache[cell_num][lineno]
        except KeyError:
            return tracer
        if safety.store_history and logger.getEffectiveLevel() <= logging.WARNING:
            try:
                source = get_ipython().all_ns_refs[0]['In'][cell_num].strip().split('\n')
                logger.warning(' %3d: %9s >>> %s', lineno, event, source[lineno-1])
            except (KeyError, IndexError) as e:
                logger.error('%s: cell %d, line %d', e, cell_num, lineno)

        trace_stmt = state.traced_statements.get(
            id(stmt_node),
            TraceStatement(safety, frame, stmt_node, state.cur_frame_scope)
        )
        state.traced_statements[id(stmt_node)] = trace_stmt
        state.state_transition_hook(event, trace_stmt)
        return tracer
    return tracer



# -*- coding: utf-8 -*-
import ast
import builtins
from contextlib import contextmanager
import logging
from typing import cast, TYPE_CHECKING

from ..data_symbol import DataSymbol, DataSymbolType
from ..scope import NamespaceScope
from ..utils import retrieve_namespace_attr_or_sub

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Set, Tuple, Union
    from ..safety import DependencySafety
    from ..scope import Scope
    DeepRef = Tuple[int, Optional[str], Tuple[str, ...]]
    Mutation = Tuple[int, Tuple[str, ...]]
    RefCandidate = Optional[Tuple[int, int, Optional[str]]]
    SavedStoreData = Tuple[NamespaceScope, Any, str, bool]
    TextualCallNestingStackFrame = Tuple[Scope, bool]
    TextualCallNestingStack = List[TextualCallNestingStackFrame]
    AttrSubStackFrame = Tuple[
        List[SavedStoreData], Set[DeepRef], Set[Mutation], RefCandidate, Set[str], Scope, Scope, TextualCallNestingStack
    ]
    AttrSubStack = List[AttrSubStackFrame]

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class AttrSubTracingManager(object):
    def __init__(self, safety: 'DependencySafety',
                 active_scope: 'Scope', trace_event_counter: 'List[int]'):
        self.safety = safety
        self.original_active_scope = active_scope
        self.active_scope = active_scope
        self.trace_event_counter = trace_event_counter
        self.attrsub_tracer_name = '_NBSAFETY_ATTR_TRACER'
        self.end_tracer_name = '_NBSAFETY_ATTR_TRACER_END'
        self.arg_recorder_name = '_NBSAFETY_ARG_RECORDER'
        self.scope_pusher_name = '_NBSAFETY_SCOPE_PUSHER'
        self.scope_popper_name = '_NBSAFETY_SCOPE_POPPER'
        setattr(builtins, self.attrsub_tracer_name, self.attrsub_tracer)
        setattr(builtins, self.end_tracer_name, self.end_tracer)
        setattr(builtins, self.arg_recorder_name, self.arg_recorder)
        setattr(builtins, self.scope_pusher_name, self.scope_pusher)
        setattr(builtins, self.scope_popper_name, self.scope_popper)
        self.ast_transformer = AttrSubTracingNodeTransformer(
            self.attrsub_tracer_name, self.end_tracer_name, self.arg_recorder_name,
            self.scope_pusher_name, self.scope_popper_name,
        )
        self.loaded_data_symbols: Set[DataSymbol] = set()
        self.saved_store_data: List[SavedStoreData] = []
        self.deep_ref_candidate: RefCandidate = None
        self.deep_refs: Set[DeepRef] = set()
        self.mutations: Set[Mutation] = set()
        self.recorded_args: Set[str] = set()
        self.nested_call_stack: TextualCallNestingStack = []
        self.stack: AttrSubStack = []
        self._waiting_for_call = False

    @property
    def active_scope_for_call(self) -> 'Scope':
        if self._waiting_for_call:
            return self.nested_call_stack[-1][0]
        return self.active_scope

    def __del__(self):
        if hasattr(builtins, self.attrsub_tracer_name):
            delattr(builtins, self.attrsub_tracer_name)
        if hasattr(builtins, self.end_tracer_name):
            delattr(builtins, self.end_tracer_name)
        if hasattr(builtins, self.arg_recorder_name):
            delattr(builtins, self.arg_recorder_name)
        if hasattr(builtins, self.scope_pusher_name):
            delattr(builtins, self.scope_pusher_name)
        if hasattr(builtins, self.scope_popper_name):
            delattr(builtins, self.scope_popper_name)

    def push_stack(self, new_scope: 'Scope'):
        self.stack.append((
            self.saved_store_data,
            self.deep_refs,
            self.mutations,
            self.deep_ref_candidate,
            self.recorded_args,
            self.active_scope,
            self.original_active_scope,
            self.nested_call_stack,
        ))
        self.saved_store_data = []
        self.deep_refs = set()
        self.mutations = set()
        self.recorded_args = set()
        self.original_active_scope = new_scope
        self.active_scope = new_scope
        self.nested_call_stack = []

    def pop_stack(self):
        (
            self.saved_store_data,
            self.deep_refs,
            self.mutations,
            self.deep_ref_candidate,
            self.recorded_args,
            self.active_scope,
            self.original_active_scope,
            self.nested_call_stack,
        ) = self.stack.pop()

    @staticmethod
    def debug_attribute_tracer(obj, attr, ctx):
        logger.debug('%s attr %s of obj %s', ctx, attr, obj)
        return obj

    def attrsub_tracer(self, obj, attr_or_subscript, is_subscript, ctx, call_context, obj_name=None):
        if obj is None:
            return None
        if not isinstance(attr_or_subscript, (str, int)):
            return obj
        obj_id = id(obj)
        scope = self.safety.namespaces.get(obj_id, None)
        # print('%s attr %s of obj %s' % (ctx, attr, obj))
        if scope is None:
            class_scope = self.safety.namespaces.get(id(obj.__class__), None)
            if class_scope is not None and not is_subscript:
                # print('found class scope %s containing %s' % (class_scope, list(class_scope.all_data_symbols_this_indentation())))
                scope = class_scope.clone(obj)
                if obj_name is not None:
                    scope.scope_name = obj_name
                self.safety.namespaces[obj_id] = scope
            else:
                # print('no scope for class', obj.__class__)
                # if self.safety.trace_state.prev_trace_stmt.finished:
                #     # avoid creating new scopes if we already did this computation
                #     self.active_scope = None
                #     return obj
                try:
                    scope_name = next(iter(self.safety.aliases.get(obj_id, None))).name if obj_name is None else obj_name
                except (TypeError, StopIteration):
                    scope_name = '<unknown namespace>'

                # FIXME: brittle strategy for determining parent scope of obj
                if (
                    obj_name is not None and
                    obj_name not in self.safety.trace_state.prev_trace_stmt_in_cur_frame.frame.f_locals
                ):
                    parent_scope = self.safety.global_scope
                else:
                    parent_scope = self.active_scope
                scope = NamespaceScope(obj, self.safety, scope_name, parent_scope=parent_scope)
                self.safety.namespaces[obj_id] = scope
        self.active_scope = scope
        # if scope is None:  # or self.safety.trace_state.prev_trace_stmt.finished:
        #     if ctx in ('Store', 'AugStore'):
        #         self.active_scope = self.original_active_scope
        #     return obj
        if scope is None or self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished:
            return obj
        elif ctx in ('Store', 'AugStore') and scope is not None:
            self.saved_store_data.append((scope, obj, attr_or_subscript, is_subscript))
            # reset active scope here
            self.active_scope = self.original_active_scope
        if ctx == 'Load':
            # save off event counter and obj_id
            # if event counter didn't change when we process the Call retval, and if the
            # retval is None, this is a likely signal that we have a mutation
            # TODO: this strategy won't work if the arguments themselves lead to traced function calls
            # print('looking for', attr_or_subscript)
            data_sym = scope.lookup_data_symbol_by_name_this_indentation(
                attr_or_subscript, is_subscript=is_subscript
            )
            if data_sym is None:
                try:
                    obj_attr_or_sub = retrieve_namespace_attr_or_sub(obj, attr_or_subscript, is_subscript)
                    symbol_type = DataSymbolType.SUBSCRIPT if is_subscript else DataSymbolType.DEFAULT
                    data_sym = DataSymbol(attr_or_subscript, symbol_type, obj_attr_or_sub, scope, self.safety, refresh_cached_obj=True)
                    # this is to prevent refs to the scope object from being considered as stale if we just load it
                    data_sym.defined_cell_num = data_sym.required_cell_num = scope.max_defined_timestamp
                    scope.put(attr_or_subscript, data_sym)
                    # print('put', data_sym, 'in', scope.full_namespace_path)
                    # FIXME: DataSymbols should probably register themselves with the alias manager at creation
                    self.safety.aliases[id(obj_attr_or_sub)].add(data_sym)
                except (AttributeError, KeyError, IndexError):
                    pass
            self.deep_ref_candidate = None
            if call_context:
                self.deep_ref_candidate = (self.trace_event_counter[0], obj_id, obj_name)
            elif data_sym is not None:
                # TODO: if we have a.b.c, will this consider a.b loaded as well as a.b.c? This is bad if so.
                self.loaded_data_symbols.add(data_sym)
        return obj

    def end_tracer(self, obj):
        if self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished:
            self.active_scope = self.original_active_scope
            return obj
        if self.deep_ref_candidate is not None:
            evt_counter, obj_id, obj_name = self.deep_ref_candidate
            self.deep_ref_candidate = None
            if evt_counter == self.trace_event_counter[0]:
                if obj is None:
                    self.mutations.add((obj_id, tuple(self.recorded_args)))
                else:
                    self.deep_refs.add((obj_id, obj_name, tuple(self.recorded_args)))
        # print('reset active scope from', self.active_scope, 'to', self.original_active_scope)
        self.active_scope = self.original_active_scope
        self.recorded_args = set()
        return obj

    def arg_recorder(self, obj, name):
        if self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished:
            return obj
        self.recorded_args.add(name)
        return obj

    def scope_pusher(self, obj):
        # if self.safety.trace_state.prev_trace_stmt.finished:
        #     return obj
        self.nested_call_stack.append((self.active_scope, self._waiting_for_call))
        self._waiting_for_call = True
        self.active_scope = self.original_active_scope
        return obj

    def scope_popper(self, obj):
        # if self.safety.trace_state.prev_trace_stmt.finished:
        #     return obj
        self.active_scope, self._waiting_for_call = self.nested_call_stack.pop()
        return obj

    def stmt_transition_hook(self):
        self._waiting_for_call = False

    def reset(self):
        self.loaded_data_symbols = set()
        self.saved_store_data = []
        self.deep_refs = set()
        self.mutations = set()
        self.deep_ref_candidate = None
        self.active_scope = self.original_active_scope
        self.nested_call_stack = []
        self.stmt_transition_hook()


class AttrSubTracingNodeTransformer(ast.NodeTransformer):
    def __init__(self, attrsub_tracer: str, end_tracer: str, arg_recorder: str, scope_pusher: str, scope_popper: str):
        self.attrsub_tracer = attrsub_tracer
        self.end_tracer = end_tracer
        self.arg_recorder = arg_recorder
        self.scope_pusher = scope_pusher
        self.scope_popper = scope_popper
        self.inside_attrsub_load_chain = False

    @contextmanager
    def attrsub_load_context(self, override=True):
        old = self.inside_attrsub_load_chain
        self.inside_attrsub_load_chain = override
        yield
        self.inside_attrsub_load_chain = old

    def visit_Attribute(self, node: 'ast.Attribute', call_context=False):
        return self.visit_Attribute_or_Subscript(node, call_context)

    def visit_Subscript(self, node: 'ast.Subscript', call_context=False):
        return self.visit_Attribute_or_Subscript(node, call_context)

    def visit_Attribute_or_Subscript(self, node: 'Union[ast.Attribute, ast.Subscript]', call_context=False):
        is_load = isinstance(node.ctx, ast.Load)
        is_subscript = isinstance(node, ast.Subscript)
        # TODO: expand beyond simple slices
        if is_subscript:
            sub_node = cast(ast.Subscript, node)
            if isinstance(sub_node.slice, ast.Index):
                attr_or_sub = sub_node.slice.value
            else:
                logger.debug('unimpled slice: %s', sub_node.slice)
                return node
            # elif isinstance(sub_node.slice, ast.Slice):
            #     raise ValueError('unimpled slice: %s' % sub_node.slice)
            # elif isinstance(sub_node.slice, ast.ExtSlice):
            #     raise ValueError('unimpled slice: %s' % sub_node.slice)
            # else:
            #     raise ValueError('unexpected slice: %s' % sub_node.slice)
        else:
            attr_node = cast(ast.Attribute, node)
            attr_or_sub = ast.Str(attr_node.attr)

        extra_args = []
        if isinstance(node.value, ast.Name):
            extra_args = [ast.Str(node.value.id)]

        with self.attrsub_load_context():
            replacement_value = ast.Call(
                func=ast.Name(self.attrsub_tracer, ast.Load()),
                args=[
                    self.visit(node.value),
                    attr_or_sub,
                    ast.NameConstant(is_subscript),
                    ast.Str(node.ctx.__class__.__name__),
                    ast.NameConstant(call_context),
                ] + extra_args,
                keywords=[]
            )
        ast.copy_location(replacement_value, node.value)
        node.value = replacement_value
        new_node: Union[ast.Attribute, ast.Subscript, ast.Call] = node
        if not self.inside_attrsub_load_chain and is_load:
            new_node = ast.Call(
                func=ast.Name(self.end_tracer, ast.Load()),
                args=[node],
                keywords=[]
            )
        return new_node

    def visit_Call(self, node: ast.Call):
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            with self.attrsub_load_context():
                node.func = self.visit_Attribute_or_Subscript(node.func, call_context=True)

        replacement_args = []
        for arg in node.args:
            if isinstance(arg, ast.Name):
                replacement_args.append(cast(ast.expr, ast.Call(
                    func=ast.Name(self.arg_recorder, ast.Load()),
                    args=[arg, ast.Str(arg.id)],
                    keywords=[]
                )))
                ast.copy_location(replacement_args[-1], arg)
            else:
                with self.attrsub_load_context(False):
                    replacement_args.append(self.visit(arg))
        node.args = replacement_args
        replacement_kwargs = []
        for kwarg in node.keywords:
            if isinstance(kwarg.value, ast.Name):
                new_kwarg_value = cast(ast.expr, ast.Call(
                    func=ast.Name(self.arg_recorder, ast.Load()),
                    args=[kwarg.value, ast.Str(kwarg.value.id)],
                    keywords=[]
                ))
                ast.copy_location(new_kwarg_value, kwarg.value)
                kwarg.value = new_kwarg_value
                replacement_kwargs.append(kwarg)
            else:
                with self.attrsub_load_context(False):
                    replacement_kwargs.append(self.visit(kwarg))
        node.keywords = replacement_kwargs

        # in order to ensure that the args are processed with appropriate active scope,
        # we need to push current active scope before processing the args and pop after
        # (pop happens on function return as opposed to in tracer)
        node.func = ast.Call(
            func=ast.Name(self.scope_pusher, ast.Load()),
            args=[node.func],
            keywords=[],
        )

        node = ast.Call(
            func=ast.Name(self.scope_popper, ast.Load()),
            args=[node],
            keywords=[]
        )

        if self.inside_attrsub_load_chain:
            return node

        replacement_node = ast.Call(
            func=ast.Name(self.end_tracer, ast.Load()),
            args=[node],
            keywords=[]
        )
        ast.copy_location(replacement_node, node)
        return replacement_node



# -*- coding: utf-8 -*-
from enum import Enum
import logging
from typing import cast, TYPE_CHECKING
import weakref

from .ipython_utils import cell_counter
from .utils import retrieve_namespace_attr_or_sub

if TYPE_CHECKING:
    from typing import Any, Optional, Set, Union
    from .safety import DependencySafety
    from .scope import Scope, NamespaceScope

logger = logging.getLogger(__name__)

NOT_FOUND = object()


class DataSymbolType(Enum):
    DEFAULT = 'default'
    SUBSCRIPT = 'subscript'
    FUNCTION = 'function'
    CLASS = 'class'


class DataSymbol(object):
    def __init__(
            self,
            name: 'Union[str, int]',
            symbol_type: 'DataSymbolType',
            obj: 'Any',
            containing_scope: 'Scope',
            safety: 'DependencySafety',
            parents: 'Optional[Set[DataSymbol]]' = None,
            refresh_cached_obj=False,
    ):
        # print(containing_scope, name, obj, is_subscript)
        self.name = name
        self.symbol_type = symbol_type
        tombstone, obj_ref, has_weakref = self._update_obj_ref_inner(obj)
        self._tombstone = tombstone
        self._obj_ref = obj_ref
        self._has_weakref = has_weakref
        self.cached_obj_ref = None
        self._cached_has_weakref = None
        self.cached_obj_id = None
        self.cached_obj_type = None
        if refresh_cached_obj:
            self._refresh_cached_obj()
        self.containing_scope = containing_scope
        self.safety = safety
        if parents is None:
            parents = set()
        self.parents: Set[DataSymbol] = parents
        self.children: Set[DataSymbol] = set()
        self.readable_name = containing_scope.make_namespace_qualified_name(self)

        if self.is_function:
            self.call_scope = self.containing_scope.make_child_scope(self.name)
        else:
            self.call_scope = None

        self.defined_cell_num = cell_counter()

        # The notebook cell number this is required to have to not be considered stale
        self.required_cell_num = self.defined_cell_num

        self.fresher_ancestors: Set[DataSymbol] = set()
        self.namespace_data_syms_with_stale: Set[DataSymbol] = set()

        # Will never be stale if no_warning is True
        self.disable_warnings = False

    def __repr__(self):
        return f'<{self.readable_name}>'

    def __str__(self):
        return self.readable_name

    def __hash__(self):
        return hash((self.full_path, self.is_subscript))

    @property
    def is_subscript(self):
        return self.symbol_type == DataSymbolType.SUBSCRIPT

    @property
    def is_class(self):
        return self.symbol_type == DataSymbolType.CLASS

    @property
    def is_function(self):
        return self.symbol_type == DataSymbolType.FUNCTION

    def _get_obj(self) -> 'Any':
        if self._has_weakref:
            return self._obj_ref()
        else:
            return self._obj_ref

    def _get_cached_obj(self) -> 'Any':
        if self._cached_has_weakref:
            return self.cached_obj_ref()
        else:
            return self.cached_obj_ref

    def shallow_clone(self, new_obj, new_containing_scope, symbol_type):
        return self.__class__(self.name, symbol_type, new_obj, new_containing_scope, self.safety)

    @property
    def obj_id(self):
        return id(self._get_obj())

    @property
    def obj_type(self):
        return type(self._get_obj())

    @property
    def namespace(self):
        return self.safety.namespaces.get(self.obj_id, None)

    @property
    def full_path(self):
        return self.containing_scope.full_path + (self.name,)

    @property
    def full_namespace_path(self):
        return self.containing_scope.make_namespace_qualified_name(self)

    @property
    def is_garbage(self):
        return (
            self._tombstone
            or self.containing_scope.is_garbage
            or not self.containing_scope.is_globally_accessible
            or (self._has_weakref and self._get_obj() is None)
        )

    def _obj_reference_expired_callback(self, *_):
        # just write a tombstone here; we'll do a batch collect after the main part of the cell is done running
        # can potentially support GC in the background further down the line
        self._tombstone = True

    def collect_self_garbage(self):
        for parent in self.parents:
            parent.children.discard(self)
        for child in self.children:
            child.parents.discard(self)
        self_aliases = self.safety.aliases.get(self.cached_obj_id, None)
        if self_aliases is not None:
            self_aliases.discard(self)
            if len(self_aliases) == 0:
                # kill the alias but leave the namespace
                # namespace needs to stick around to properly handle the staleness propagation protocol
                self.safety.aliases.pop(self.cached_obj_id, None)

    def update_type(self, new_type):
        self.symbol_type = new_type
        if self.is_function:
            self.call_scope = self.containing_scope.make_child_scope(self.name)
        else:
            self.call_scope = None

    def update_obj_ref(self, obj):
        tombstone, obj_ref, has_weakref = self._update_obj_ref_inner(obj)
        self._tombstone = tombstone
        self._obj_ref = obj_ref
        self._has_weakref = has_weakref

    def _update_obj_ref_inner(self, obj):
        tombstone = False
        try:
            obj_ref = weakref.ref(obj, self._obj_reference_expired_callback)
            has_weakref = True
        except TypeError:
            obj_ref = obj
            has_weakref = False
        return tombstone, obj_ref, has_weakref

    def _refresh_cached_obj(self):
        self.cached_obj_ref = self._obj_ref
        self.cached_obj_id = self.obj_id
        self.cached_obj_type = self.obj_type
        self._cached_has_weakref = self._has_weakref

    def update_deps(
        self,
        new_deps: 'Set[DataSymbol]',
        overwrite=True,
        mutated=False,
    ):

        # quick last fix to avoid ovewriting if we appear inside the set of deps to add
        overwrite = overwrite and self not in new_deps
        new_deps.discard(self)
        if overwrite:
            for parent in self.parents - new_deps:
                parent.children.discard(self)
            self.parents = set()

        for new_parent in new_deps - self.parents:
            if new_parent is None:
                continue
            new_parent.children.add(self)
            self.parents.add(new_parent)

        self.required_cell_num = -1
        self._propagate_update(self._get_obj(), self, set(), set(), refresh=True, mutated=mutated)
        self._refresh_cached_obj()
        self.safety.updated_symbols.add(self)

    def refresh(self):
        self.fresher_ancestors = set()
        self.defined_cell_num = cell_counter()
        self.required_cell_num = self.defined_cell_num
        self.namespace_data_syms_with_stale = set()
        self._propagate_refresh_to_namespace_parents(set())

    def _propagate_update_to_deps(
            self,
            updated_dep: 'DataSymbol',
            seen: 'Set[DataSymbol]',
            parent_seen: 'Set[DataSymbol]',
    ):
        # print(self, 'propagate to child deps', self.children)
        if self.should_mark_stale(updated_dep):
            self.fresher_ancestors.add(updated_dep)
            self.required_cell_num = cell_counter()
        for child in self.children:
            child._propagate_update(child._get_obj(), updated_dep, seen, parent_seen)

    def _propagate_update(
            self, new_parent_obj: 'Any',
            updated_dep: 'DataSymbol', seen, parent_seen,
            refresh=False, mutated=False
    ):
        # look at old obj_id and cur obj_id
        # walk down namespace hierarchy from old obj_id, and track corresponding DCs from cur obj_id
        # a few cases to consider:
        # 1. mismatched obj ids or unavailable from new hierarchy:
        #    mark old dc as mutated AND stale, and propagate to old dc children
        # 2. new / old DataSymbols have same obj ids:
        #    mark old dc as mutated, but NOT stale, and propagate to children
        #    Q: should old dc additionally be refreshed?
        #    Technically it should already be fresh, since if it's still a descendent of this namespace, we probably
        #    had to ref the namespace ancestor, which should have been caught by the checker if the descendent has
        #    some other stale ancestor. If not fresh, let's mark it so and log a warning about a potentially stale usage
        if self._tombstone or self in seen:
            return
        seen.add(self)
        # print('propagate update from', self)

        new_id = None if new_parent_obj is NOT_FOUND else id(new_parent_obj)

        if updated_dep is self:
            old_parent_obj = self._get_cached_obj()
            old_id = self.cached_obj_id
        else:
            old_parent_obj = self._get_obj()
            old_id = self.obj_id

        namespace = self.safety.namespaces.get(old_id, None)
        if namespace is None and refresh:  # if we are at a leaf
            self._propagate_update_to_namespace_parents(updated_dep, seen, parent_seen, refresh=refresh)
        for dc in [] if namespace is None else namespace.all_data_symbols_this_indentation():
            dc_in_self_namespace = False
            if new_parent_obj is NOT_FOUND:
                dc._propagate_update(NOT_FOUND, updated_dep, seen, parent_seen, refresh=refresh, mutated=mutated)
            else:
                try:
                    obj = self._get_obj()
                    obj_attr_or_sub = retrieve_namespace_attr_or_sub(obj, dc.name, dc.is_subscript)
                    dc._propagate_update(
                        obj_attr_or_sub, updated_dep, seen, parent_seen, refresh=refresh, mutated=mutated
                    )
                    dc_in_self_namespace = True
                    if new_parent_obj is not old_parent_obj and new_id is not None:
                        new_namespace = self.safety.namespaces.get(new_id, None)
                        if new_namespace is None:
                            new_namespace = namespace.shallow_clone(new_parent_obj)
                            self.safety.namespaces[new_id] = new_namespace
                        # TODO: handle class data cells properly;
                        #  in fact; we still need to handle aliases of class data cells
                        if dc.name not in new_namespace.data_symbol_by_name(dc.is_subscript):
                            new_dc = dc.shallow_clone(obj_attr_or_sub, new_namespace, dc.symbol_type)
                            new_namespace.put(dc.name, new_dc)
                            self.safety.updated_symbols.add(new_dc)
                except (KeyError, IndexError, AttributeError):
                    dc._propagate_update(NOT_FOUND, updated_dep, seen, parent_seen, refresh=refresh, mutated=mutated)
            if dc_in_self_namespace and dc.has_stale_ancestor:
                if dc.should_mark_stale(updated_dep) and self.should_mark_stale(updated_dep):
                    self.namespace_data_syms_with_stale.add(dc)
            else:
                self.namespace_data_syms_with_stale.discard(dc)

        if mutated or self.cached_obj_id != self.obj_id:
            self._propagate_update_to_deps(updated_dep, seen, parent_seen)

        if updated_dep is self:
            return

        if refresh:
            for alias in self.safety.aliases[old_id]:
                if alias.defined_cell_num < alias.required_cell_num < cell_counter():
                    logger.debug('possible stale usage of namespace descendent %s' % alias)
                if len(alias.namespace_data_syms_with_stale) > 0:
                    logger.debug('unexpected stale namespace symbols for symbol %s: %s' % (alias, alias.namespace_data_syms_with_stale))
                    alias.namespace_data_syms_with_stale.clear()
                if old_id != new_id or mutated:
                    # TODO: better equality testing
                    #  Doing equality testing properly requires that we still have a reference to old object around;
                    #  we should be using weak refs, which complicates this.
                    #  Depth also makes this challenging.
                    alias._propagate_update_to_deps(updated_dep, seen, parent_seen)
                # TODO: this will probably work, but will also unnecessarily propagate the refresh up to
                #  namespace parents. Need a mechanism to mark for refresh without parent propagation.
                self.safety.updated_symbols.add(alias)
        else:
            for alias in self.safety.aliases[old_id]:
                # print('propagate', updated_dep, 'to', alias, 'via', self, updated_dep.defined_cell_num, alias.defined_cell_num, self.defined_cell_num)
                alias._propagate_update_to_deps(updated_dep, seen, parent_seen)
            if namespace is None and self.should_mark_stale(updated_dep):  # if we are at a leaf
                self._propagate_update_to_namespace_parents(updated_dep, seen, parent_seen, refresh=refresh)

    def mark_mutated(self):
        # print(self, 'mark mutated')
        self.update_deps(set(), overwrite=False, mutated=True)

    def _propagate_refresh_to_namespace_parents(self, seen):
        if not self.containing_scope.is_namespace_scope or self in seen:
            return
        seen.add(self)
        containing_scope: 'NamespaceScope' = cast('NamespaceScope', self.containing_scope)
        if containing_scope.max_defined_timestamp == cell_counter():
            return
        containing_scope.max_defined_timestamp = cell_counter()
        containing_namespace_obj_id = containing_scope.obj_id
        for alias in self.safety.aliases[containing_namespace_obj_id]:
            alias.namespace_data_syms_with_stale.discard(self)
            if not alias.has_stale_ancestor:
                alias.fresher_ancestors = set()
                alias._propagate_refresh_to_namespace_parents(seen)

    def _propagate_update_to_namespace_parents(self, updated_dep, seen, parent_seen, refresh):
        if not self.containing_scope.is_namespace_scope or self in parent_seen:
            return
        parent_seen.add(self)
        containing_scope = cast('NamespaceScope', self.containing_scope)
        containing_namespace_obj_id = containing_scope.obj_id
        for alias in self.safety.aliases[containing_namespace_obj_id]:
            # print('propagate from ns parent', alias, 'with refresh=', refresh)
            if refresh and not self.has_stale_ancestor:
                alias.namespace_data_syms_with_stale.discard(self)
                if not alias.has_stale_ancestor:
                    alias.fresher_ancestors = set()
            alias._propagate_update_to_namespace_parents(updated_dep, seen, parent_seen, refresh)
            for alias_child in alias.children:
                if alias_child.obj_id == containing_namespace_obj_id:
                    continue
                # Next, complicated check to avoid propagating along a class -> instance edge.
                # The only time this is OK is when we changed the class, which will not be the case here.
                alias_child_namespace = alias_child.namespace
                if alias_child_namespace is not None:
                    if alias_child_namespace.cloned_from is containing_scope:
                        if updated_dep.namespace is not containing_scope:
                            continue
                alias_child._propagate_update(alias_child._get_obj(), updated_dep, seen, parent_seen)
            if refresh:
                # containing_scope.max_defined_timestamp = cell_counter()
                self.safety.updated_scopes.add(containing_scope)
            else:
                if self.should_mark_stale(updated_dep):
                    alias.namespace_data_syms_with_stale.add(self)

    @property
    def has_stale_ancestor(self):
        if self.disable_warnings:
            return False
        return self.defined_cell_num < self.required_cell_num or len(self.namespace_data_syms_with_stale) > 0

    def should_mark_stale(self, updated_dep):
        if self.disable_warnings:
            return False
        if updated_dep is self:
            return False
        should_mark_stale = not self.safety.no_stale_propagation_for_same_cell_definition
        should_mark_stale = should_mark_stale or updated_dep.defined_cell_num != self.defined_cell_num
        return should_mark_stale



# -*- coding: utf-8 -*-
from ipykernel.ipkernel import IPythonKernel
from ..version import __version__
from ..safety import DependencySafety
# import warnings
#Avoid networkx gives a FutureWarning
# TODO (smacke): this disables all FutureWarnings; disable until we can figure out a better system
# warnings.filterwarnings("ignore", category=FutureWarning)


class SafeKernel(IPythonKernel):
    implementation = 'kernel'
    implementation_version = __version__

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._safety = DependencySafety(use_comm=True)

    def do_execute(self, code, silent, store_history=False, user_expressions=None, allow_stdin=False):
        super_ = super()

        def _run_cell_func(cell):
            return super_.do_execute(cell, silent, store_history, user_expressions, allow_stdin)
        return self._safety.safe_execute(code, _run_cell_func)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Set
    from ..data_symbol import DataSymbol


class NoUpdateYet(object):
    def update(self, dep_update: 'DependencyUpdate') -> 'DependencyUpdate':
        return dep_update


class DependencyUpdate(object):
    """Just records the metadata needed to update deps for a DataSymbol."""
    def __init__(
        self,
        deps: 'Set[DataSymbol]',
        overwrite: bool,
        mutate: bool = False
    ):
        self.deps = deps
        self.overwrite = overwrite
        self.mutate = mutate

    def update(self, dep_update: 'DependencyUpdate') -> 'DependencyUpdate':
        if dep_update.overwrite:
            dep_update.mutate = dep_update.mutate or self.mutate
            return dep_update
        else:
            self.mutate = self.mutate or dep_update.mutate
            self.deps |= dep_update.deps
            return self



# -*- coding: utf-8 -*-
import ast

from nbsafety.analysis.attr_symbols import get_attrsub_symbol_chain, CallPoint as Cp


def test_basic():
    node = ast.parse('a.b.c.d.e.f').body[0].value
    symchain = get_attrsub_symbol_chain(node)
    assert symchain.symbols == ('a', 'b', 'c', 'd', 'e', 'f')
    assert symchain.call_points == ()


def test_calls_none_at_endpoints():
    node = ast.parse('a.b.c().d.e().f').body[0].value
    symchain = get_attrsub_symbol_chain(node)
    assert symchain.symbols == ('a', 'b', Cp('c'), 'd', Cp('e'), 'f')
    assert symchain.call_points == (Cp('c'), Cp('e'))


def test_calls_at_endpoints():
    node = ast.parse('a().b.c().d.e.f()').body[0].value
    symchain = get_attrsub_symbol_chain(node)
    assert symchain.symbols == (Cp('a'), 'b', Cp('c'), 'd', 'e', Cp('f'))
    assert symchain.call_points == (Cp('a'), Cp('c'), Cp('f'))


def test_hash():
    symchain_set = set()
    symchain_set.add(get_attrsub_symbol_chain('f.read()'))
    symchain_set.add(get_attrsub_symbol_chain('f.read'))
    symchain_set.add('f')
    symchain_set.add(('f', 'read'))
    assert len(symchain_set) == 4



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
import json
import os
import sys
from setuptools import setup, find_packages
from tempfile import TemporaryDirectory

import versioneer

pkg_name = 'nbsafety'

DISPLAY_NAME = f"Python 3 ({pkg_name})"
KERNEL_JSON = {
    "argv": [
        sys.executable, "-m", "nbsafety.kernel", "-f", "{connection_file}",
    ],
    "display_name": DISPLAY_NAME,
    "language": "python",
    "codemirror_mode": "shell",
}


def read_file(fname):
    with open(fname, 'r', encoding='utf8') as f:
        return f.read()


history = read_file('HISTORY.rst')
requirements = read_file('requirements.txt').strip().split()

if sys.argv[1] == 'sdist':
    should_dump_kernel_json = False
    tempdir_or_nullcontext = contextlib.nullcontext
else:
    should_dump_kernel_json = True
    tempdir_or_nullcontext = TemporaryDirectory
with tempdir_or_nullcontext() as td:
    extra_data_files = []
    if should_dump_kernel_json:
        os.chmod(td, 0o755)  # Starts off as 700, not user readable
        kernel_json_file = os.path.join(td, 'kernel.json')
        with open(kernel_json_file, 'w') as f:
            json.dump(KERNEL_JSON, f, sort_keys=True)
        extra_data_files.append(("share/jupyter/kernels/nbsafety", [os.path.relpath(kernel_json_file, os.curdir)]))
    setup(
        name=pkg_name,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        author='Stephen Macke',
        author_email='stephen.macke@gmail.com',
        description='Fearless interactivity for Jupyter notebooks.',
        long_description=read_file('README.md'),
        long_description_content_type='text/markdown',
        url='https://github.com/nbsafety-project/nbsafety',
        packages=find_packages(exclude=[
            'binder',
            'docs',
            'scratchspace',
            'notebooks',
            'img',
            'test',
            'scripts',
            'markdown',
            'versioneer.py',
            'frontend',
            'blueprint.json',
        ]),
        include_package_data=True,
        data_files=[
            # like `jupyter nbextension install --sys-prefix`
            ("share/jupyter/nbextensions/nbsafety", [
                "nbsafety/resources/nbextension/index.js",
            ]),
            ("share/jupyter/nbextensions/nbsafety", [
                "nbsafety/resources/nbextension/index.js.map",
            ]),
            # like `jupyter nbextension enable --sys-prefix`
            ("etc/jupyter/nbconfig/notebook.d", [
                "nbsafety/resources/nbextension/nbsafety.json",
            ]),
            # like `python -m nbsafety.install --sys-prefix`
            ("share/jupyter/kernels/nbsafety", [
                "nbsafety/resources/kernel/logo-32x32.png",
            ]),
            ("share/jupyter/kernels/nbsafety", [
                "nbsafety/resources/kernel/logo-64x64.png",
            ]),
        ] + extra_data_files,
        install_requires=requirements,
        license='BSD-3-Clause',
        zip_safe=False,
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
        ],
    )

# python setup.py sdist bdist_wheel --universal
# twine upload dist/*



# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
import logging
from typing import TYPE_CHECKING

from .dep_update import DependencyUpdate
from ..analysis import get_statement_symbol_edges
from ..utils import retrieve_namespace_attr_or_sub

if TYPE_CHECKING:
    from types import FrameType
    from typing import List, Optional, Set
    from ..data_symbol import DataSymbol
    from ..safety import DependencySafety
    from ..scope import Scope, NamespaceScope

logger = logging.getLogger(__name__)


class TraceStatement(object):
    def __init__(self, safety: 'DependencySafety', frame: 'FrameType', stmt_node: 'ast.stmt', scope: 'Scope'):
        self.safety = safety
        self.frame = frame
        self.stmt_node = stmt_node
        self.scope = scope
        self.class_scope: Optional[NamespaceScope] = None
        self.call_point_deps: List[Set[DataSymbol]] = []
        self._marked_finished = False

    @contextmanager
    def replace_active_scope(self, new_active_scope):
        old_scope = self.scope
        self.scope = new_active_scope
        yield
        self.scope = old_scope

    @property
    def finished(self):
        return self._marked_finished
        # return self.marked_finished and isinstance(self.stmt_node, (ast.For, ast.Lambda))

    def compute_rval_dependencies(self, rval_symbol_refs=None):
        if rval_symbol_refs is None:
            symbol_edges, _ = get_statement_symbol_edges(self.stmt_node)
            if len(symbol_edges) == 0:
                rval_symbol_refs = set()
            else:
                rval_symbol_refs = set.union(*symbol_edges.values()) - {None}
        rval_data_symbols = set()
        for name in rval_symbol_refs:
            if name is None:
                continue
            maybe_rval_dsym = self.scope.lookup_data_symbol_by_name(name)
            if maybe_rval_dsym is not None:
                rval_data_symbols.add(maybe_rval_dsym)
        return rval_data_symbols.union(*self.call_point_deps) | self.safety.attr_trace_manager.loaded_data_symbols

    def get_post_call_scope(self, old_scope: 'Scope'):
        if isinstance(self.stmt_node, ast.ClassDef):
            # classes need a new scope before the ClassDef has finished executing,
            # so we make it immediately
            return self.scope.make_child_scope(self.stmt_node.name, obj_id=-1)

        if not isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            # TODO: probably the right thing is to check is whether a lambda appears somewhere inside the ast node
            # if not isinstance(self.ast_node, ast.Lambda):
            #     raise TypeError('unexpected type for ast node %s' % self.ast_node)
            return old_scope
        func_name = self.stmt_node.name
        func_cell = self.scope.lookup_data_symbol_by_name(func_name)
        if func_cell is None:
            # TODO: brittle; assumes any user-defined and traceable function will always be present; is this safe?
            return old_scope
        if not func_cell.is_function:
            raise TypeError('got non-function symbol %s for name %s' % (func_cell.full_path, func_name))
        return func_cell.call_scope

    def _gather_lval_data_symbol_dep_updates(self):
        symbol_edges, should_overwrite = get_statement_symbol_edges(self.stmt_node)
        deep_rval_deps = self._gather_deep_ref_rval_dsyms()
        is_function_def = isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef))
        is_class_def = isinstance(self.stmt_node, ast.ClassDef)
        if is_function_def or is_class_def:
            assert len(symbol_edges) == 1
            # assert not lval_symbol_refs.issubset(rval_symbol_refs)
        for lval_name, rval_names in symbol_edges.items():
            if lval_name is None:
                continue
            should_overwrite_for_name = should_overwrite and lval_name not in rval_names
            rval_deps = self.compute_rval_dependencies(rval_symbol_refs=rval_names - {lval_name}) | deep_rval_deps
            # print('create edges from', rval_deps, 'to', lval_name, should_overwrite_for_name)
            if is_class_def:
                assert self.class_scope is not None
                class_ref = self.frame.f_locals[self.stmt_node.name]
                class_obj_id = id(class_ref)
                self.class_scope.obj_id = class_obj_id
                self.safety.namespaces[class_obj_id] = self.class_scope
            # if is_function_def:
            #     print('create function', name, 'in scope', self.scope)
            try:
                obj = self.frame.f_locals[lval_name]
                dep_update = DependencyUpdate(rval_deps, should_overwrite_for_name)
                dsym = self.scope.upsert_data_symbol_for_name(
                    lval_name, obj, False, is_function_def=is_function_def, class_scope=self.class_scope,
                )
                self.safety.dep_updates[dsym] = self.safety.dep_updates[dsym].update(dep_update)
            except KeyError:
                pass
        if len(symbol_edges) == 0:
            rval_deps = deep_rval_deps
        else:
            rval_deps = self.compute_rval_dependencies(
                rval_symbol_refs=set.union(*symbol_edges.values()) - {None}
            ) | deep_rval_deps
        for scope, obj, attr_or_sub, is_subscript in self.safety.attr_trace_manager.saved_store_data:
            # print(scope, obj, attr_or_sub, is_subscript)
            try:
                attr_or_sub_obj = retrieve_namespace_attr_or_sub(obj, attr_or_sub, is_subscript)
            except (AttributeError, KeyError, IndexError):
                continue
            should_overwrite = not isinstance(self.stmt_node, ast.AugAssign)
            scope_to_use = scope.get_earliest_ancestor_containing(id(attr_or_sub_obj), is_subscript)
            if scope_to_use is None:
                # Nobody before `scope` has it, so we'll insert it at this level
                scope_to_use = scope
            old_dsym = scope_to_use.lookup_data_symbol_by_name_this_indentation(attr_or_sub)
            if old_dsym is not None and scope_to_use.is_globally_accessible:
                if attr_or_sub not in scope_to_use.data_symbol_by_name(is_subscript):
                    # In this case, we are copying from a class and we need the dsym from which we are copying
                    # as able to propagate to the new dsym.
                    # Example:
                    # class Foo:
                    #     shared = 99
                    # foo = Foo()
                    # foo.shared = 42  # old_dc refers to Foo.shared here
                    # Earlier, we were explicitly adding Foo.shared as a dependency of foo.shared as follows:
                    # deps.add(old_dc)
                    # But it turns out not to be necessary because foo depends on Foo, and changing Foo.shared will
                    # propagate up the namespace hierarchy to Foo, which propagates to foo, which then propagates to
                    # all of foo's namespace children (e.g. foo.shared).
                    # This raises the question of whether we should draw the foo <-> Foo edge, since irrelevant namespace
                    # children could then also be affected (e.g. some instance variable foo.x).
                    # Perhaps a better strategy is to prevent propagation along this edge unless class Foo is redeclared.
                    # If we do this, then we should go back to explicitly adding the dep as follows:
                    # EDIT: added check to avoid propagating along class -> instance edge when class not redefined, so now
                    # it is important to explicitly add this dep.
                    # deps.add(old_dc)
                    rval_deps.add(old_dsym)
            dep_update = DependencyUpdate(rval_deps, should_overwrite)
            dsym = scope_to_use.upsert_data_symbol_for_name(
                attr_or_sub, attr_or_sub_obj, is_subscript, is_function_def=False, class_scope=None
            )
            self.safety.dep_updates[dsym] = self.safety.dep_updates[dsym].update(dep_update)

    def _gather_deep_ref_rval_dsyms(self):
        deep_ref_rval_dsyms = set()
        for deep_ref_obj_id, deep_ref_name, deep_ref_args in self.safety.attr_trace_manager.deep_refs:
            deep_ref_arg_dsyms = set(self.scope.lookup_data_symbol_by_name(arg) for arg in deep_ref_args) - {None}
            deep_ref_rval_dsyms |= deep_ref_arg_dsyms
            if deep_ref_name is None:
                deep_ref_rval_dsyms |= self.safety.aliases.get(deep_ref_obj_id, set())
            else:
                deep_ref_dc = self.scope.lookup_data_symbol_by_name(deep_ref_name)
                if deep_ref_dc is not None and deep_ref_dc.obj_id == deep_ref_obj_id:
                    deep_ref_rval_dsyms.add(deep_ref_dc)
                else:
                    deep_ref_rval_dsyms |= self.safety.aliases.get(deep_ref_obj_id, set())
        return deep_ref_rval_dsyms

    def handle_dependencies(self):
        if not self.safety.dependency_tracking_enabled:
            return
        for mutated_obj_id, mutation_args in self.safety.attr_trace_manager.mutations:
            mutation_arg_dsyms = set(self.scope.lookup_data_symbol_by_name(arg) for arg in mutation_args) - {None}
            dep_update = DependencyUpdate(mutation_arg_dsyms, False, mutate=True)
            self.safety.dep_updates[mutated_obj_id] = self.safety.dep_updates[mutated_obj_id].update(dep_update)
        if self.has_lval:
            self._gather_lval_data_symbol_dep_updates()
        else:
            if len(self.safety.attr_trace_manager.saved_store_data) > 0:
                logger.warning('saw unexpected state in saved_store_data: %s',
                               self.safety.attr_trace_manager.saved_store_data)

    def finished_execution_hook(self):
        if self.finished:
            return
        # print('finishing stmt', self.stmt_node)
        self._marked_finished = True
        self.handle_dependencies()
        self.safety.attr_trace_manager.reset()

    @property
    def has_lval(self):
        # TODO: expand to method calls, etc.
        return isinstance(self.stmt_node, (
            ast.Assign, ast.AnnAssign, ast.AugAssign, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef, ast.For
        ))
